#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <getopt.h>
#include <immintrin.h>
#include <inttypes.h>
#include <limits.h>
#include <linux/if_packet.h>
#include <math.h>
#include <netdb.h>
#include <netinet/ether.h>
#include <netinet/in.h>
#include <net/if.h>
#include <pthread.h>
#include <sched.h>
#include <semaphore.h>
#include <signal.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

typedef struct {
  char name[64];
  char value[256];
} param_t;

param_t param_list[512];
int param_list_len;

char * _trx_get_param_string(void *app_opaque, const char *prop_name) {
  for(int i = 0; i < param_list_len; i++) {
    if(strcmp(prop_name, param_list[i].name) == 0) {
      return param_list[i].value;
    }
  }
  return NULL;
}
int _trx_get_param_double(void *app_opaque, double *pval,
                            const char *prop_name) {
  for(int i = 0; i < param_list_len; i++) {
    if(strcmp(prop_name, param_list[i].name) == 0) {
      *pval = strtod(param_list[i].value, NULL);
      return 0;
    }
  }
  return 1;
}

#include "trx-ecpri/trx_ecpri_dpdk.c"

TRXState s1;
TRXDriverParams params;

static void enb(void);

static int search(char * p, char * s) {
  int k = 0;
  for(int i = 0; s[i] != '\0'; i++) {
    if(s[i] == p[k]) {
      k++;
      if(k == strlen(p))
        return 1;
    } else {
      i -= k;
      k = 0;
    }
  }
  return 0;
}

static int tx_test = 0;
static int rx_test = 0;
static int freq;
static int period_ns;

int main(int argc, char * argv[]) {
  (void) argc;
  (void) argv;
  FILE * f;
  char * line = NULL;
  size_t len = 0;
  ssize_t read;
  int k = 0;

  if(argc != 2) {
    fprintf(stderr, "Usage: ./eNB-simulator enb.cfg\n");
    exit(EXIT_FAILURE);
  }

  f = fopen(argv[1], "r");
  if(f == NULL)
    exit(EXIT_FAILURE);

  while((read = getline(&line, &len, f)) != -1) {
    if(search("rf_driver", line))
      break;
  }
  while((read = getline(&line, &len, f)) != -1) {
    int i, s, e;
    if(line[0] == '}')
      break;
    for(i = 0; line[i] == ' '; i++);
    s = i;
    for(; line[i] != ':' || line[i] == '\0'; i++)
      param_list[k].name[i-s] = line[i];
    i++;
    for(; line[i] == ' '; i++);
    s = i;
    e = 0;
    for(; line[i] != '\0'; i++)
      if(line[i] == ',')
        e = i;
    for(i = s; i < e; i++) {
      if(line[i] == '"') {
        s++;
        continue;
      }
      param_list[k].value[i-s] = line[i];
    }
    k++;
  }
  param_list_len = k;
  fclose(f);
  if (line)
    free(line);

  double _rx_test;
  double _tx_test;
  double _freq;
  _trx_get_param_double(NULL, &_rx_test, "rx_test");
  _trx_get_param_double(NULL, &_tx_test, "tx_test");
  _trx_get_param_double(NULL, &_freq, "frame_frequency");
  rx_test = (int) _rx_test;
  tx_test = (int) _tx_test;
  freq = (int) _freq;

  // Lock all current and future pages from preventing of being paged to
  // swap
  if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
    log_error("TRX_ECPRI", "mlockall failed");
  }

  log_info("ENB-SIMULATOR", "Starting...\n");

  s1.trx_get_param_string = &_trx_get_param_string;
  s1.trx_get_param_double = &_trx_get_param_double;
  s1.trx_api_version = 15;

  params.sample_rate[0].num = 1.0;
  params.sample_rate[0].den = 1.0;

  trx_driver_init(&s1);
  trx_ecpri_start(&s1, &params);
  enb();
}

#define NS_PER_SEC INT64_C(1000000000)

volatile int64_t first_rx_ts = -1;

#define FAST_WRITE_N_SAMPLES (59968 / 32)
float ** fast_write_tx_samples;

static void init_fast_write() {
  fast_write_tx_samples = (float**) malloc(sizeof(float*) * 4);
  for(int i = 0; i < 4; i++) {
    fast_write_tx_samples[i] = (float*) malloc(sizeof(float) * 64 * FAST_WRITE_N_SAMPLES);
  }
}

static void fast_write(int wait_read) {
  struct timespec next;
  int64_t offset = 0;
  int64_t period = (NS_PER_SEC * FAST_WRITE_N_SAMPLES / freq);

  log_info("FAST_WRITE", "period = %li", period);


  while(wait_read && first_rx_ts == -1);
  if(wait_read)
    offset = first_rx_ts;
  clock_gettime(CLOCK_TAI, &next);

  for(int64_t i = 0;; i++) {

    //add_ns(&next, period);

    s1.trx_write_func2(&s1,
          offset + i*32*FAST_WRITE_N_SAMPLES,
          (const void **) fast_write_tx_samples,
          32*FAST_WRITE_N_SAMPLES,
    0, NULL);
    //clock_nanosleep(CLOCK_TAI, TIMER_ABSTIME, &next, NULL);
  }
}

#define BURSTS 100
#define WRITE_DATA_N_SAMPLES (59968 / 32)
float ** write_data_tx_samples;
int write_data_k = 0;
float write_data_t[256] = {0.2806, 0.1765, 0.241, 0.315, 0.1102, 0.1687, 0.2796, 0.1007, 0.1532, 0.1416, 0.3251, 0.1648, 0.0943, 0.2117, 0.0819, 0.2903, 0.2501, 0.0474, 0.2062, 0.0929, 0.1164, 0.2628, 0.233, 0.0807, 0.0622, 0.1898, 0.0353, 0.0621, 0.2952, 0.3302, 0.0593, 0.0149, 0.1332, 0.1623, 0.2873, 0.2878, 0.2234, 0.2496, 0.1093, 0.1939, 0.0139, 0.2309, 0.0788, 0.0413, 0.032, 0.2459, 0.0053, 0.1913, 0.0128, 0.0753, 0.2993, 0.2474, 0.2497, 0.2482, 0.0296, 0.316, 0.0697, 0.0964, 0.1447, 0.1936, 0.2217, 0.2048, 0.1711, 0.3068, 0.2425, 0.2568, 0.1369, 0.2054, 0.2321, 0.0719, 0.2702, 0.0296, 0.2425, 0.1185, 0.0195, 0.0897, 0.2867, 0.2843, 0.2479, 0.1714, 0.2353, 0.1614, 0.121, 0.1995, 0.0721, 0.1928, 0.0733, 0.2868, 0.2046, 0.0172, 0.0027, 0.0572, 0.1601, 0.2939, 0.1993, 0.2842, 0.0202, 0.1897, 0.0031, 0.1861, 0.1305, 0.1519, 0.0283, 0.2113, 0.2693, 0.0196, 0.3043, 0.0369, 0.0326, 0.1655, 0.2832, 0.2635, 0.1708, 0.1424, 0.0898, 0.1463, 0.0282, 0.2291, 0.0158, 0.0665, 0.1167, 0.2735, 0.1775, 0.3135, 0.0885, 0.2902, 0.0881, 0.2111, 0.2576, 0.2711, 0.3058, 0.1956, 0.1731, 0.0076, 0.0491, 0.2363, 0.2494, 0.2543, 0.2979, 0.3211, 0.0391, 0.2565, 0.2005, 0.3278, 0.3223, 0.2468, 0.1193, 0.2862, 0.0784, 0.1754, 0.1152, 0.2882, 0.0257, 0.079, 0.1545, 0.1653, 0.1096, 0.2328, 0.0746, 0.1112, 0.1772, 0.0905, 0.1404, 0.2764, 0.187, 0.138, 0.2594, 0.0333, 0.2169, 0.0259, 0.1605, 0.0934, 0.3246, 0.3075, 0.1163, 0.1556, 0.1923, 0.0994, 0.1196, 0.0013, 0.1475, 0.3157, 0.2043, 0.2966, 0.1123, 0.0212, 0.2995, 0.2666, 0.1253, 0.1169, 0.1842, 0.2481, 0.0942, 0.1586, 0.0143, 0.2794, 0.1392, 0.2289, 0.1335, 0.1818, 0.302, 0.2306, 0.1953, 0.1039, 0.2353, 0.2674, 0.1898, 0.0835, 0.0386, 0.1237, 0.1454, 0.019, 0.1964, 0.248, 0.1655, 0.1459, 0.2217, 0.1022, 0.1079, 0.0025, 0.0447, 0.1059, 0.1805, 0.1608, 0.2823, 0.078, 0.1647, 0.2305, 0.0922, 0.3287, 0.0871, 0.2898, 0.3061, 0.0612, 0.046, 0.215, 0.2707, 0.3313, 0.0825, 0.2921, 0.2113, 0.2864, 0.296, 0.1408, 0.2583, 0.1971, 0.0467, 0.1458, 0.2265, 0.3308, 0.1136, 0.0656, 0.0612, 0.2644, 0.2627, 0.1968};

static void init_write_data() {


  write_data_tx_samples = (float**) malloc(sizeof(float*) * 4);
  for(int i = 0; i < 4; i++) {
    write_data_tx_samples[i] = (float*) malloc(sizeof(float) * 64 * WRITE_DATA_N_SAMPLES);
    //for(int j = 0; j < 64 * WRITE_DATA_N_SAMPLES; j++) {
    //  write_data_tx_samples[i][j] = t[i * 64 + (j % 64)];
    //}
  }
}

static void write_data(int wait_read) {
  struct timespec next;
  float ** tx_samples;
  int offset = 0;
  int64_t period_ns = NS_PER_SEC / freq;

  clock_gettime(CLOCK_TAI, &next);

  for(int i = 0;; i++) {
    add_ns(&next, period_ns * BURSTS);

    if(wait_read && first_rx_ts == -1)
      continue;
    else if(wait_read)
      offset = first_rx_ts;

    for(int l = 0; l < BURSTS; l++) {
      for(int c = 0; c < 4; c++) {
        for(int s = 0; s < 64; s++) {
          write_data_tx_samples[c][64 * l + s] = write_data_t[(c * 64 + s + write_data_k) % 256];
        }
      }
      write_data_k++;
    }

    s1.trx_write_func2(&s1,
          offset + BURSTS * i*32,
          (const void **) write_data_tx_samples,
          BURSTS * 32,
    0, NULL);

    clock_nanosleep(CLOCK_TAI, TIMER_ABSTIME, &next, NULL);
  }
}

static void write_simple_tdd(int n_ul, int n_dl, int wait_read) {
  struct timespec next;
  float ** tx_samples_ones;
  int uldl_i = 0;
  int offset = 0;
  int64_t period_ns = NS_PER_SEC / freq;

  clock_gettime(CLOCK_TAI, &next);
  tx_samples_ones = (float**) malloc(sizeof(float*) * 4);
  for(int i = 0; i < 4; i++) {
    tx_samples_ones[i] = (float*) malloc(sizeof(float) * 64);
    for(int j = 0; j < 64; j++)
      tx_samples_ones[i][j] = 1.0f;
  }

  for(int i = 0;; i++) {
    add_ns(&next, period_ns);

    if(wait_read && first_rx_ts == -1)
      continue;
    else if(wait_read)
      offset = first_rx_ts;

    uldl_i = (uldl_i + 1) % (n_ul + n_dl);

    s1.trx_write_func2(&s1,
          offset + i*32,
          uldl_i < n_ul ? NULL : (const void **) tx_samples_ones,
          32,
    0, NULL);

    clock_nanosleep(CLOCK_TAI, TIMER_ABSTIME, &next, NULL);
  }
}
static void read_periodic(void) {
  struct timespec next;
  trx_timestamp_t ptimestamp;
  float ** rx_samples;
  int64_t period_ns = NS_PER_SEC / freq;

  clock_gettime(CLOCK_TAI, &next);
  rx_samples = (float**) malloc(sizeof(float*) * 4);
  for(int i = 0; i < 4; i++)
    rx_samples[i] = (float*) malloc(sizeof(float) * 64);

  for(int i = 0;; i++) {
    add_ns(&next, period_ns);

    s1.trx_read_func2(&s1, &ptimestamp, (void **) rx_samples, 32, 0, NULL);
    first_rx_ts = i == 0 ? ptimestamp : first_rx_ts;

    clock_nanosleep(CLOCK_TAI, TIMER_ABSTIME, &next, NULL);
  }
}
static void read_all(int n) {
  trx_timestamp_t ptimestamp;
  float ** rx_samples;

  rx_samples = (float**) malloc(sizeof(float*));
  for(int i = 0; i < 4; i++)
    rx_samples[i] = (float*) malloc(sizeof(float) * 64 * n);

  for(int i = 0;; i++) {
    s1.trx_read_func2(&s1, &ptimestamp, (void **) rx_samples, 32 * n, 0, NULL);
    first_rx_ts = (i == 0) ? ptimestamp : first_rx_ts;

    usleep(100);
  }
}

static void do_nothing(void) {
  while(1)
    sleep(1);
}

static void *read_thread(void *p) {
  (void) p;

  log_info("READ_THREAD", "Thread init");

  if(rx_test)
    read_all(10);

  do_nothing();

  do_nothing();
  read_periodic();
  read_all(10000);

  pthread_exit(EXIT_SUCCESS);
}
static void *write_thread(void *p) {
  (void) p;

  log_info("READ_THREAD", "Thread init");

  if(tx_test)
    write_data(0);

  do_nothing();

  write_simple_tdd(1723, 1423, 1);
  write_data(0);
  do_nothing();
  fast_write(1);

  pthread_exit(EXIT_SUCCESS);
}


static void enb() {
  pthread_t read_pthread;
  pthread_t write_pthread;

  log_info("ENB-SIMULATOR", "Starting threads");

  if(tx_test)
    init_write_data();

  if (pthread_create(&write_pthread, NULL, write_thread, NULL))
    error(EXIT_FAILURE, errno, "Couldn't create write thread");
  if (pthread_create(&read_pthread, NULL, read_thread, NULL))
    error(EXIT_FAILURE, errno, "Couldn't create read thread");

  while(1) {
    sleep(1);
  }
}
