#!/usr/bin/env python3

from jinja2 import Template
import os
import sys
import json

server_map = {
  '003': {
    'full_name': 'Tiogapass 003',
    'log_directory': '/home/slapos/ecpri-logs',
    'nic_list': {
      'mellanox': {
        'if_list': {
          '0': {
            'if_name': 'ens9f0',
            'mac': 'b8:59:9f:07:7e:2a',
            'pci_id': '0000:5e:00.0',
          },
          '1': {
            'if_name': 'ens9f1',
            'mac': 'b8:59:9f:07:7e:2b',
            'pci_id': '0000:5e:00.1',
          },
        },
      },
      'i210': {
        'if_list': {
          '0': {
            'if_name': 'eno1',
            'mac': '00:22:4d:da:5b:a0',
            'pci_id': '0000:04:00.0',
          },
        },
      },
    },
  },
  '004': {
    'full_name': 'Tiogapass 004',
    'log_directory': '/home/slapos/ecpri-logs',
    'nic_list': {
      'mellanox': {
        'if_list': {
          '0': {
            'if_name': 'ens9f0',
            'mac': 'b8:59:9f:07:86:42',
            'pci_id': '0000:5e:00.0',
          },
          '1': {
            'if_name': 'ens9f1',
            'mac': 'b8:59:9f:07:86:43',
            'pci_id': '0000:5e:00.1',
          },
        },
      },
      'i210': {
        'if_list': {
          '0': {
            'if_name': 'eno1',
            'mac': '00:22:4d:da:8a:42',
            'pci_id': '0000:04:00.0',
          },
        },
      },
    },
  },
  'hfr001': {
    'full_name': 'HFR Tiogapass 001',
    'log_directory': '/home/hfr',
    'nic_list': {
      'mellanox': {
        'if_list': {
          '0': {
            'if_name': 'ens9f0',
            'mac': 'b8:59:9f:07:82:ca',
            'pci_id': '0000:5e:00.0',
          },
          '1': {
            'if_name': 'ens9f1',
            'mac': 'b8:59:9f:07:82:cb',
            'pci_id': '0000:5e:00.1',
          },
        },
      },
      'intel': {
        'if_list': {
          '0': {
            'if_name': 'ens1f0',
            'mac': 'b4:96:91:a7:1b:28',
            'pci_id': '0000:18:00.0',
          },
          '1': {
            'if_name': 'ens1f1',
            'mac': 'b4:96:91:a7:1b:29',
            'pci_id': '0000:18:00.1',
          },
        },
      },
      'i210': {
        'if_list': {
          '0': {
            'if_name': 'eno1',
            'mac': '00:22:4d:da:8a:38',
            'pci_id': '0000:04:00.0',
          },
        },
      },
    },
  },
  'lillem64': {
    'full_name': 'Lille Lab M6424 Switch',
    'log_directory': '',
    'nic_list': {
      '0': {
        'if_list': {
          '0': {
            'if_name': '',
            'mac': '04:09:a5:0f:9f:4a',
            'pci_id': '',
          },
        },
      },
    },
  },
  'hfrm64': {
    'full_name': 'HFR Korea Lab M6424 Switch',
    'log_directory': '',
    'nic_list': {
      '0': {
        'if_list': {
          '0': {
            'if_name': '',
            'mac': '04:09:a5:0f:76:1c',
            'pci_id': '',
          },
        },
      },
    },
  },
}

if len(sys.argv) < 8:
    print("Wrong argument count")
    exit()

params      = sys.argv[1]
rec         = sys.argv[2]
rec_nic     = sys.argv[3]
rec_if      = sys.argv[4]
re          = sys.argv[5]
re_nic      = sys.argv[6]
re_if       = sys.argv[7]

def get_eal_param(server, nic_key, iface_key):
    l = []
    for _nic_key in server_map[server]['nic_list']:
        nic = server_map[server]['nic_list'][_nic_key]
        for _iface_key in nic['if_list']:
            iface = nic['if_list'][_iface_key]
            if _nic_key != nic_key or _iface_key != iface_key:
                l.append(iface['pci_id'])
    return '-l 10,20 ' + ''.join(map(lambda x: '-b {} '.format(x), l))

context = {
    'rec_mac': server_map[rec]['nic_list'][rec_nic]['if_list'][rec_if]['mac'],
    're_mac': server_map[re]['nic_list'][re_nic]['if_list'][re_if]['mac'],
    'rec_if': server_map[rec]['nic_list'][rec_nic]['if_list'][rec_if]['if_name'],
    'log_directory': server_map[rec]['log_directory'],
    'dpdk_options': get_eal_param(rec, rec_nic, rec_if),
}

with open(params, 'r') as f:
  lines = [l for l in f]
lines = [l.split(',')[0] for l in lines]
lines = [l.split(':', 2) for l in lines]

for name, value in lines:
  context[name] = value

if len(sys.argv) >= 9:
  context["trace_offset"] = sys.argv[8]

with open('enb.jinja2.cfg', 'r') as f:
    template = Template(f.read())
print(template.render(**context))
