#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <getopt.h>
#include <immintrin.h>
#include <inttypes.h>
#include <limits.h>
#include <linux/if_packet.h>
#include <math.h>
#include <netdb.h>
#include <netinet/ether.h>
#include <netinet/in.h>
#include <net/if.h>
#include <pthread.h>
#include <sched.h>
#include <semaphore.h>
#include <signal.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include "../utils.c"

typedef struct {
  float re;
  float im;
} Complex;

#define AVX2
#ifdef AVX2
#include "../bf1_avx2.c"
#else
#include "../bf1_sse4.c"
#endif

#define N_CHANNELS 4
#define N_SAMPLES 32
//#define FREQ 3840000
#define FREQ 2000000

static Complex ** iq_samples;
static int16_t * int16_samples;
static uint8_t * out;

static int64_t diff_ts(struct timespec t1, struct timespec t2) {
  int64_t diff;
  diff = NSEC_PER_SEC * ((int)t1.tv_sec - (int)t2.tv_sec);
  diff += ((int)t1.tv_nsec - (int)t2.tv_nsec);
  return diff;
}


int main() {

  struct timespec t[5];
  int ti = 0;
  const float mult = 32767.;
  const float imult = 1. / 32767.;

  if (mlockall(MCL_CURRENT | MCL_FUTURE))
    fprintf(stderr, "mlockall failed\n");

  iq_samples = (Complex **) malloc(N_CHANNELS * sizeof(Complex *));
  for(int i = 0; i < N_CHANNELS; i++)
    iq_samples[i] = (Complex *) malloc(FREQ * N_SAMPLES * sizeof(Complex));
  int16_samples = (int16_t *) malloc(FREQ * N_SAMPLES * N_CHANNELS * 2 * sizeof(int16_t));
  out = (uint8_t *) malloc(60 * FREQ * N_CHANNELS * sizeof(uint8_t));
  for(int i = 0; i < N_CHANNELS ; i++)
    for(int j = 0; j < FREQ * N_SAMPLES * 2; j++)
      ((float *) iq_samples[i])[j] = 0.4f * random() / ((float) RAND_MAX);

  clock_gettime(CLOCK_MONOTONIC, &t[ti++]);
#ifndef AVX2
  for(int i = 0; i < N_CHANNELS ; i++)
    float_to_int16(int16_samples + i * (FREQ * N_SAMPLES * 2), (float *) iq_samples[i], FREQ * N_SAMPLES * 2, mult);
#endif

  clock_gettime(CLOCK_MONOTONIC, &t[ti++]);
#ifdef AVX2
  for(int j = 0; j < FREQ; j++)
    for(int i = 0; i < N_CHANNELS ; i++)
      encode_s64_b60_2(out + i * 60 + j * N_CHANNELS * 60, ((float *) (iq_samples[i] + j * N_SAMPLES)));
#else
  for(int i = 0; i < FREQ * N_CHANNELS; i++)
    encode_bf1(out + i * 60, int16_samples + i * N_SAMPLES * 2);
#endif

  clock_gettime(CLOCK_MONOTONIC, &t[ti++]);
#ifdef AVX2
  for(int j = 0; j < FREQ; j++)
    for(int i = 0; i < N_CHANNELS ; i++)
      decode_s64_b60_2((float *) (iq_samples[i] + j * N_SAMPLES), out + i * 60 + j * N_CHANNELS * 60);
#else
  for(int i = 0; i < FREQ * N_CHANNELS; i++)
    decode_bf1(int16_samples + i * N_SAMPLES * 2, out + i * 60, 16);
#endif

  clock_gettime(CLOCK_MONOTONIC, &t[ti++]);
#ifndef AVX2
  for(int i = 0; i < N_CHANNELS ; i++)
    int16_to_float((float *) iq_samples[i], int16_samples + i * (FREQ * N_SAMPLES * 2), FREQ * N_SAMPLES * 2, imult);
#endif

  clock_gettime(CLOCK_MONOTONIC, &t[ti++]);

  for(int i = 1; i < ti; i++) {
    int64_t diff = diff_ts(t[i], t[i - 1]);
    int64_t ms = diff / 1000000;
    int64_t us = (diff - ms * 1000000) / 1000;
    int64_t ns = diff - ms * 1000000 - us * 1000;
    printf("%d: %li.%li.%li\n", i, ms, us, ns);
  }
  
  return 0;
}
